==> dovecot-0.9.tar.gz: dovecot-0.9/src/login/client.c <==
/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "iobuffer.h"
#include "client.h"
#include "auth.h"

#include <syslog.h>

/* Disconnect client after idling this many seconds */
#define CLIENT_LOGIN_IDLE_TIMEOUT 60

static HashTable *clients;
static Timeout to_idle;

static void client_input(void *user_data, int fd, IO io);

static void login_callback(Client *client, const char *user, int success)
{
	if (success)
		client_destroy(client, t_strconcat("LOGIN ", user, NULL));
	else {
		client_syslog(client, t_strconcat("Failed LOGIN ", user, NULL));
		client_send_tagline(client, "NO Login failed.");

		/* client is allowed to send data again */
		client->io = io_add(client->fd, IO_READ, client_input, client);
	}
}

static void authenticate_callback(Client *client, const char *name, int success)
{
	if (success)
		client_destroy(client, t_strconcat("AUTHENTICATE ",
						   name, NULL));
	else {
		client_syslog(client, t_strconcat("Failed AUTHENTICATE ",
						  name, NULL));
		client_send_tagline(client, "NO Authentication failed.");

		/* client is allowed to send data again */
		client->io = io_add(client->fd, IO_READ, client_input, client);
	}
}

static int client_login(Client *client, char *const *args)
{
	/* <user> <password> */
	if (args[0] == NULL || args[1] == NULL)
		return FALSE;

	if (default_auth != NULL && default_auth->login != NULL) {
		/* don't read any input from client until login is finished */
		io_remove(client->io);
		client->io = NULL;

		default_auth->login(client, args[0], args[1], login_callback);
		return TRUE;
	}

	client_send_tagline(client, "NO Login failed.");
	return TRUE;
}

static int client_authenticate(Client *client, char *const *args)
{
	AuthModule *auth;

	/* <authentication mechanism name> */
	if (args[0] == NULL)
		return FALSE;

	auth = auth_find(args[0]);
	if (auth != NULL && auth->authenticate != NULL) {
		/* don't read any input from client until
		   authentication is finished */
		io_remove(client->io);
		client->io = NULL;

		auth->authenticate(client, authenticate_callback);
		return TRUE;
	}

	client_send_tagline(client, "NO Authentication failed.");
	return TRUE;
}

static int client_capability(Client *client, char *const *args)
{
	client_send_line(client, "* CAPABILITY " CAPABILITY_STRING);
	client_send_tagline(client, "OK Capability completed.");
	return TRUE;
}

static int client_noop(Client *client, char *const *args)
{
	client_send_tagline(client, "OK NOOP completed.");
	return TRUE;
}

static int client_logout(Client *client, char *const *args)
{
	client_send_line(client, "* BYE Logging out");
	client_send_tagline(client, "OK Logout completed.");
	client_destroy(client, "Logged out");
	return TRUE;
}

static void client_input(void *user_data, int fd, IO io)
{
	Client *client = user_data;
	char *const *args;
	unsigned char *line;

	client->last_input = ioloop_time;

	i_free(client->tag);
	client->tag = i_strdup("*");

	switch (io_buffer_read(client->inbuf)) {
	case -2:
		/* buffer full */
		client_send_line(client, "* BYE Input buffer full, aborting");
		client_destroy(client, "Input buffer full");
		return;
	case -1:
		/* disconnected */
		client_destroy(client, "Disconnected");
		return;
	default:
		/* something was read */
	}

	while ((line = io_buffer_next_line(client->inbuf)) != NULL) {
		/* split the arguments, make sure we have at
		   least tag + command */
		args = t_strsplit(line, " ");
		if (args[0] != NULL) {
			i_free(client->tag);
			client->tag = i_strdup(args[0]);
		}

		if (strarray_length(args) >= 2) {
			if (strcasecmp(args[1], "LOGIN") == 0) {
				if (client_login(client, args+2))
					return;
			} else if (strcasecmp(args[1], "AUTHENTICATE") == 0) {
				if (client_authenticate(client, args+2))
					return;
			} else if (strcasecmp(args[1], "CAPABILITY") == 0) {
				if (client_capability(client, args+2))
					return;
			} else if (strcasecmp(args[1], "NOOP") == 0) {
				if (client_noop(client, args+2))
					return;
			} else if (strcasecmp(args[1], "LOGOUT") == 0) {
				if (client_logout(client, args+2))
					return;
			}
		}

		/* error */
		client_send_tagline(client, "BAD Error in IMAP command "
				    "received by server.");
	}
}

Client *client_create(int fd, IPADDR *ip)
{
	Client *client;

	client = i_new(Client, 1);
	client->refcount = 1;

	memcpy(&client->ip, ip, sizeof(IPADDR));
	client->fd = fd;
	client->io = io_add(fd, IO_READ, client_input, client);
	client->inbuf = io_buffer_create(fd, default_pool,
					 IO_PRIORITY_DEFAULT, 8192);
	client->outbuf = io_buffer_create(fd, default_pool,
					  IO_PRIORITY_DEFAULT, 1024);
        client->last_input = ioloop_time;
	hash_insert(clients, client, client);

	client_send_line(client, "* OK " PACKAGE " ready.");
	return client;
}

void client_destroy(Client *client, const char *reason)
{
	if (reason != NULL)
		client_syslog(client, reason);

	hash_remove(clients, client);

	io_buffer_close(client->inbuf);
	io_buffer_close(client->outbuf);

	if (client->io != NULL) {
		io_remove(client->io);
		client->io = NULL;
	}

	net_disconnect(client->fd);
	client->fd = -1;

	client_unref(client);
}

void client_ref(Client *client)
{
	client->refcount++;
}

int client_unref(Client *client)
{
	if (--client->refcount > 0)
		return TRUE;

	io_buffer_destroy(client->inbuf);
	io_buffer_destroy(client->outbuf);

	i_free(client);
	return FALSE;
}

void client_send_line(Client *client, const char *line)
{
	io_buffer_send(client->outbuf, line, strlen(line));
	io_buffer_send(client->outbuf, "\r\n", 2);
}

void client_send_tagline(Client *client, const char *line)
{
	client_send_line(client, t_strconcat(client->tag, " ", line, NULL));
}

void client_syslog(Client *client, const char *text)
{
	char host[MAX_IP_LEN];

	if (net_ip2host(&client->ip, host) == -1)
		host[0] = '\0';

	syslog(LOG_INFO, "%s [%s]", text, host);
}

static void client_hash_check_idle(void *key, void *value, void *user_data)
{
	Client *client = key;

	if (ioloop_time - client->last_input >= CLIENT_LOGIN_IDLE_TIMEOUT) {
		client_send_line(client, "* BYE Disconnected for inactivity.");
		client_destroy(client, "Inactivity");
	}
}

static void idle_timeout(void *user_data, Timeout timeout)
{
	hash_foreach(clients, client_hash_check_idle, NULL);
}

void clients_init(void)
{
	clients = hash_create(default_pool, 128, NULL, NULL);
	to_idle = timeout_add(1000, idle_timeout, NULL);
}

static void client_hash_destroy(void *key, void *value, void *user_data)
{
	client_destroy(key, NULL);
}

void clients_deinit(void)
{
	hash_foreach(clients, client_hash_destroy, NULL);
	hash_destroy(clients);

	timeout_remove(to_idle);
}

==> dovecot-0.91.tar.gz: dovecot-0.91/src/login/client.c <==
/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "iobuffer.h"
#include "client.h"
#include "client-authenticate.h"

#include <syslog.h>

/* Disconnect client after idling this many seconds */
#define CLIENT_LOGIN_IDLE_TIMEOUT 60

static HashTable *clients;
static Timeout to_idle;

static int cmd_capability(Client *client, char *const *args)
{
	const char *capability;

	capability = t_strconcat("* CAPABILITY " CAPABILITY_STRING, " ",
				 client_authenticate_get_capabilities(), NULL);
	client_send_line(client, capability);
	client_send_tagline(client, "OK Capability completed.");
	return TRUE;
}

static int cmd_noop(Client *client, char *const *args)
{
	client_send_tagline(client, "OK NOOP completed.");
	return TRUE;
}

static int cmd_logout(Client *client, char *const *args)
{
	client_send_line(client, "* BYE Logging out");
	client_send_tagline(client, "OK Logout completed.");
	client_destroy(client, "Logged out");
	return TRUE;
}

int client_read(Client *client)
{
	switch (io_buffer_read(client->inbuf)) {
	case -2:
		/* buffer full */
		client_send_line(client, "* BYE Input buffer full, aborting");
		client_destroy(client, "Disconnected: Input buffer full");
		return FALSE;
	case -1:
		/* disconnected */
		client_destroy(client, "Disconnected");
		return FALSE;
	default:
		/* something was read */
		return TRUE;
	}
}

static int client_command_execute(Client *client, char *const *args)
{
	str_ucase(args[0]);

	if (strcmp(args[0], "LOGIN") == 0)
		return cmd_login(client, args+1);
	if (strcmp(args[0], "AUTHENTICATE") == 0)
		return cmd_authenticate(client, args+1);
	if (strcmp(args[0], "CAPABILITY") == 0)
		return cmd_capability(client, args+1);
	if (strcmp(args[0], "NOOP") == 0)
		return cmd_noop(client, args+1);
	if (strcmp(args[0], "LOGOUT") == 0)
		return cmd_logout(client, args+1);

	return FALSE;
}

void client_input(void *user_data, int fd, IO io)
{
	Client *client = user_data;
	char *const *args;
	unsigned char *line;

	client->last_input = ioloop_time;

	i_free(client->tag);
	client->tag = i_strdup("*");

	if (!client_read(client))
		return;

	while ((line = io_buffer_next_line(client->inbuf)) != NULL) {
		/* split the arguments, make sure we have at
		   least tag + command */
		args = t_strsplit(line, " ");
		if (args[0] != NULL) {
			i_free(client->tag);
			client->tag = i_strdup(args[0]);
			args++;
		}

		if (args[0] == NULL || !client_command_execute(client, args)) {
			/* error */
			client_send_tagline(client, "BAD Error in IMAP command "
					    "received by server.");
		}
	}
}

Client *client_create(int fd, IPADDR *ip)
{
	Client *client;

	client = i_new(Client, 1);
	client->refcount = 1;

	memcpy(&client->ip, ip, sizeof(IPADDR));
	client->fd = fd;
	client->io = io_add(fd, IO_READ, client_input, client);
	client->inbuf = io_buffer_create(fd, default_pool,
					 IO_PRIORITY_DEFAULT, 8192);
	client->outbuf = io_buffer_create(fd, default_pool,
					  IO_PRIORITY_DEFAULT, 1024);
        client->last_input = ioloop_time;
	hash_insert(clients, client, client);

	client_send_line(client, "* OK " PACKAGE " ready.");
	return client;
}

void client_destroy(Client *client, const char *reason)
{
	if (reason != NULL)
		client_syslog(client, reason);

	hash_remove(clients, client);

	io_buffer_close(client->inbuf);
	io_buffer_close(client->outbuf);

	if (client->io != NULL) {
		io_remove(client->io);
		client->io = NULL;
	}

	net_disconnect(client->fd);
	client->fd = -1;

	client_unref(client);
}

void client_ref(Client *client)
{
	client->refcount++;
}

int client_unref(Client *client)
{
	if (--client->refcount > 0)
		return TRUE;

	io_buffer_destroy(client->inbuf);
	io_buffer_destroy(client->outbuf);

	i_free(client->tag);
	i_free(client->plain_login);
	i_free(client);
	return FALSE;
}

void client_send_line(Client *client, const char *line)
{
	io_buffer_send(client->outbuf, line, strlen(line));
	io_buffer_send(client->outbuf, "\r\n", 2);
}

void client_send_tagline(Client *client, const char *line)
{
	client_send_line(client, t_strconcat(client->tag, " ", line, NULL));
}

void client_syslog(Client *client, const char *text)
{
	char host[MAX_IP_LEN];

	if (net_ip2host(&client->ip, host) == -1)
		host[0] = '\0';

	syslog(LOG_INFO, "%s [%s]", text, host);
}

static void client_hash_check_idle(void *key, void *value, void *user_data)
{
	Client *client = key;

	if (ioloop_time - client->last_input >= CLIENT_LOGIN_IDLE_TIMEOUT) {
		client_send_line(client, "* BYE Disconnected for inactivity.");
		client_destroy(client, "Disconnected: Inactivity");
	}
}

static void idle_timeout(void *user_data, Timeout timeout)
{
	hash_foreach(clients, client_hash_check_idle, NULL);
}

void clients_init(void)
{
	clients = hash_create(default_pool, 128, NULL, NULL);
	to_idle = timeout_add(1000, idle_timeout, NULL);
}

static void client_hash_destroy(void *key, void *value, void *user_data)
{
	client_destroy(key, NULL);
}

void clients_deinit(void)
{
	hash_foreach(clients, client_hash_destroy, NULL);
	hash_destroy(clients);

	timeout_remove(to_idle);
}
